'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fiber = require('@react-three/fiber');
var React = require('react');
var mergeRefs = require('react-merge-refs');
var three = require('three');
var CustomShaderMaterial = require('three-custom-shader-material');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefault(React);
var mergeRefs__default = /*#__PURE__*/_interopDefault(mergeRefs);
var CustomShaderMaterial__default = /*#__PURE__*/_interopDefault(CustomShaderMaterial);

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };
  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function _objectWithoutProperties(source, excluded) {
  if (source == null) return {};
  var target = _objectWithoutPropertiesLoose(source, excluded);
  var key, i;

  if (Object.getOwnPropertySymbols) {
    var sourceSymbolKeys = Object.getOwnPropertySymbols(source);

    for (i = 0; i < sourceSymbolKeys.length; i++) {
      key = sourceSymbolKeys[i];
      if (excluded.indexOf(key) >= 0) continue;
      if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue;
      target[key] = source[key];
    }
  }

  return target;
}

var ParticlesContext = /*#__PURE__*/React.createContext(null);
var useParticles = function useParticles() {
  return React.useContext(ParticlesContext);
};

function getValue(fov) {
  return typeof fov === "function" ? fov() : fov;
}

var _excluded$3 = ["count", "setup", "continuous"];
var Emitter = /*#__PURE__*/React__default["default"].forwardRef(function (_ref, ref) {
  var _ref$count = _ref.count,
      count = _ref$count === void 0 ? 0 : _ref$count,
      setup = _ref.setup,
      _ref$continuous = _ref.continuous,
      continuous = _ref$continuous === void 0 ? false : _ref$continuous,
      props = _objectWithoutProperties(_ref, _excluded$3);

  var _useParticles = useParticles(),
      spawnParticle = _useParticles.spawnParticle;

  var object = React.useRef(null);
  React.useEffect(function () {
    if (continuous) return;
    spawnParticle(getValue(count), setup, object.current);
  }, []);
  fiber.useFrame(function () {
    if (continuous) {
      spawnParticle(getValue(count), setup, object.current);
    }
  });
  return /*#__PURE__*/React__default["default"].createElement("object3D", _extends({}, props, {
    ref: mergeRefs__default["default"]([ref, object])
  }));
});

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    enumerableOnly && (symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    })), keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = null != arguments[i] ? arguments[i] : {};
    i % 2 ? ownKeys(Object(source), !0).forEach(function (key) {
      _defineProperty(target, key, source[key]);
    }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)) : ownKeys(Object(source)).forEach(function (key) {
      Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
    });
  }

  return target;
}

function mutateShader(target, source) {
  Object.assign(target, addShaders(target, source));
}
function addShaders(a, b) {
  return {
    uniforms: _objectSpread2(_objectSpread2({}, a.uniforms), b.uniforms),
    varyings: _objectSpread2(_objectSpread2({}, a.varyings), b.varyings),
    attributes: _objectSpread2(_objectSpread2({}, a.attributes), b.attributes),
    config: _objectSpread2(_objectSpread2({}, a.config), b.config),
    vertexHeader: a.vertexHeader + b.vertexHeader,
    vertexMain: a.vertexMain + b.vertexMain,
    fragmentHeader: a.fragmentHeader + b.fragmentHeader,
    fragmentMain: a.fragmentMain + b.fragmentMain,
    resetConfig: function resetConfig() {
      var _a$resetConfig, _b$resetConfig;

      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      (_a$resetConfig = a.resetConfig) === null || _a$resetConfig === void 0 ? void 0 : _a$resetConfig.call.apply(_a$resetConfig, [a].concat(args));
      (_b$resetConfig = b.resetConfig) === null || _b$resetConfig === void 0 ? void 0 : _b$resetConfig.call.apply(_b$resetConfig, [b].concat(args));
    },
    applyConfig: function applyConfig() {
      var _a$applyConfig, _b$applyConfig;

      for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
      }

      (_a$applyConfig = a.applyConfig) === null || _a$applyConfig === void 0 ? void 0 : _a$applyConfig.call.apply(_a$applyConfig, [a].concat(args));
      (_b$applyConfig = b.applyConfig) === null || _b$applyConfig === void 0 ? void 0 : _b$applyConfig.call.apply(_b$applyConfig, [b].concat(args));
    },
    update: function update() {
      var _a$update, _b$update;

      for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        args[_key3] = arguments[_key3];
      }

      (_a$update = a.update) === null || _a$update === void 0 ? void 0 : _a$update.call.apply(_a$update, [a].concat(args));
      (_b$update = b.update) === null || _b$update === void 0 ? void 0 : _b$update.call.apply(_b$update, [b].concat(args));
    }
  };
}
/* TODO: properly type this */

function combineShaders(shaders) {
  return shaders.reduce(addShaders);
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  var _i = arr == null ? null : typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"];

  if (_i == null) return;
  var _arr = [];
  var _n = true;
  var _d = false;

  var _s, _e;

  try {
    for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

var formatValue = function formatValue(value) {
  return typeof value === "number" ? value.toFixed(5) : value;
};

function compileVariable(qualifier, name, variable) {
  var initializeValue = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : true;
  var valueString = initializeValue && "value" in variable && variable.value !== undefined ? " = ".concat(formatValue(variable.value)) : "";
  return "".concat(qualifier, " ").concat(variable.type, " ").concat(name).concat(valueString, ";");
}

function compileVariables(qualifier, variables) {
  var initializeValue = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;
  return Object.entries(variables).map(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        name = _ref2[0],
        variable = _ref2[1];

    return compileVariable(qualifier, name, variable, initializeValue);
  }).join("\n");
}

function compileProgram(shader, what) {
  return "\n    ".concat(compileVariables("uniform", shader.uniforms, false), "\n    ").concat(what === "vertex" ? compileVariables("attribute", shader.attributes, false) : "", "\n    ").concat(compileVariables("varying", shader.varyings), "\n    ").concat(shader["".concat(what, "Header")], "\n\n    void main() {\n      ").concat(shader["".concat(what, "Main")], "\n    }");
}

function compileShader(shader) {
  return {
    uniforms: shader.uniforms,
    vertexShader: compileProgram(shader, "vertex"),
    fragmentShader: compileProgram(shader, "fragment"),
    update: shader.update
  };
}

function createShader(input) {
  return _objectSpread2({
    uniforms: {},
    varyings: {},
    attributes: {},
    config: {},
    vertexHeader: "",
    vertexMain: "",
    fragmentHeader: "",
    fragmentMain: ""
  }, input);
}

var itemSizes = {
  "float": 1,
  vec2: 2,
  vec3: 3,
  vec4: 4,
  mat3: 9,
  mat4: 16,
  sampler2D: 1,
  samplerCube: 1,
  bool: 1
};
var prepareInstancedMesh = function prepareInstancedMesh(imesh, shader, maxInstanceCount) {
  /* Create all the attributes configured in the shader. */
  for (var name in shader.attributes) {
    var itemSize = itemSizes[shader.attributes[name].type];
    var buffer = new Float32Array(maxInstanceCount * itemSize);
    var attribute = new three.InstancedBufferAttribute(buffer, itemSize);
    imesh.geometry.setAttribute(name, attribute);
  }
  /* Start with an instance count of zero. */


  imesh.count = 0;
};

function animateColors () {
  var fun = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "v_progress";
  var config = {
    color: {
      min: new three.Color(),
      max: new three.Color()
    },
    alpha: {
      min: 1,
      max: 1
    }
  };
  return createShader({
    attributes: {
      color0: {
        type: "vec4"
      },
      color1: {
        type: "vec4"
      }
    },
    varyings: {
      v_color0: {
        type: "vec4"
      },
      v_color1: {
        type: "vec4"
      }
    },
    vertexMain: "\n      v_color0 = color0;\n      v_color1 = color1;\n    ",
    fragmentMain: "\n      /* Get diffuse color */\n      vec4 diffuse4 = vec4(diffuse, 1.0);\n\n      /* Apply the diffuse color */\n      csm_DiffuseColor = mix(diffuse4 * v_color0, diffuse4 * v_color1, ".concat(fun, ");\n\n      /* Mix in the texture */\n      #ifdef USE_MAP\n        csm_DiffuseColor *= texture2D(map, vUv);\n      #endif\n    "),
    config: config,
    resetConfig: function resetConfig(mesh) {
      config.color.min.setRGB(1, 1, 1);
      config.color.max.setRGB(1, 1, 1);
      config.alpha.min = 1;
      config.alpha.max = 1;
    },
    applyConfig: function applyConfig(_ref, cursor) {
      var attributes = _ref.geometry.attributes;

      /* Set color */
      attributes.color0.setXYZW(cursor, config.color.min.r, config.color.min.g, config.color.min.b, config.alpha.min);
      attributes.color1.setXYZW(cursor, config.color.max.r, config.color.max.g, config.color.max.b, config.alpha.max);
    }
  });
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) return _arrayLikeToArray(arr);
}

function _iterableToArray(iter) {
  if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter);
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
}

var tmpMatrix4 = new three.Matrix4();
var tmpScale = new three.Vector3();
function animateMovement () {
  var config = {
    position: new three.Vector3(),
    quaternion: new three.Quaternion(),
    velocity: new three.Vector3(),
    acceleration: new three.Vector3()
  };
  return createShader({
    attributes: {
      velocity: {
        type: "vec3"
      },
      acceleration: {
        type: "vec3"
      }
    },
    vertexMain: "\n      /* Apply velocity */\n      csm_Position += vec3(v_age * velocity) * mat3(instanceMatrix);\n\n      /* Apply acceleration */\n      csm_Position += vec3(0.5 * v_age * v_age * acceleration) * mat3(instanceMatrix);\n    ",
    config: config,
    resetConfig: function resetConfig(mesh) {
      config.position.set(0, 0, 0);
      config.quaternion.set(0, 0, 0, 1);
      config.velocity.set(0, 0, 0);
      config.acceleration.set(0, 0, 0);
    },
    applyConfig: function applyConfig(mesh, cursor) {
      var _attributes$velocity, _attributes$accelerat;

      /* Set origin position of the instance */
      mesh.setMatrixAt(cursor, tmpMatrix4.compose(config.position, config.quaternion, tmpScale.setScalar(1)));
      /* Set attributes */

      var attributes = mesh.geometry.attributes;

      (_attributes$velocity = attributes.velocity).setXYZ.apply(_attributes$velocity, [cursor].concat(_toConsumableArray(config.velocity.toArray())));

      (_attributes$accelerat = attributes.acceleration).setXYZ.apply(_attributes$accelerat, [cursor].concat(_toConsumableArray(config.acceleration.toArray())));
    }
  });
}

function animateScale () {
  var fun = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "v_progress";
  var config = {
    scale: {
      min: new three.Vector3(),
      max: new three.Vector3()
    }
  };
  return createShader({
    attributes: {
      scale0: {
        type: "vec3"
      },
      scale1: {
        type: "vec3"
      }
    },
    vertexMain: "\n      csm_Position *= mix(scale0, scale1, ".concat(fun, ");\n    "),
    config: config,
    resetConfig: function resetConfig(mesh) {
      config.scale.min.setScalar(1);
      config.scale.max.setScalar(1);
    },
    applyConfig: function applyConfig(_ref, cursor) {
      var _attributes$scale, _attributes$scale2;

      var attributes = _ref.geometry.attributes;

      (_attributes$scale = attributes.scale0).setXYZ.apply(_attributes$scale, [cursor].concat(_toConsumableArray(config.scale.min.toArray())));

      (_attributes$scale2 = attributes.scale1).setXYZ.apply(_attributes$scale2, [cursor].concat(_toConsumableArray(config.scale.max.toArray())));
    }
  });
}

function billboarding () {
  return createShader({
    vertexHeader: "\n      vec3 billboard(vec2 v, mat4 view){\n        vec3 up = vec3(view[0][1], view[1][1], view[2][1]);\n        vec3 right = vec3(view[0][0], view[1][0], view[2][0]);\n        vec3 p = right * v.x + up * v.y;\n        return p;\n      }\n    ",
    vertexMain: "\n      csm_Position = billboard(csm_Position.xy, viewMatrix);\n    "
  });
}

function provideDepthTexture (depthTexture) {
  return createShader({
    uniforms: {
      u_depth: {
        type: "sampler2D",
        value: depthTexture
      }
    }
  });
}

/* Big thanks to https://github.com/glslify/glsl-easings for the code below! */
var functions =
/*glsl*/
"\n  #ifndef HALF_PI\n  #define HALF_PI 1.5707963267948966\n  #endif\n\n  float backInOut(float t) {\n    float f = t < 0.5\n      ? 2.0 * t\n      : 1.0 - (2.0 * t - 1.0);\n\n    float g = pow(f, 3.0) - f * sin(f * PI);\n\n    return t < 0.5\n      ? 0.5 * g\n      : 0.5 * (1.0 - g) + 0.5;\n  }\n\n  float backIn(float t) {\n    return pow(t, 3.0) - t * sin(t * PI);\n  }\n\n  float backOut(float t) {\n    float f = 1.0 - t;\n    return 1.0 - (pow(f, 3.0) - f * sin(f * PI));\n  }\n\n  float bounceOut(float t) {\n    const float a = 4.0 / 11.0;\n    const float b = 8.0 / 11.0;\n    const float c = 9.0 / 10.0;\n\n    const float ca = 4356.0 / 361.0;\n    const float cb = 35442.0 / 1805.0;\n    const float cc = 16061.0 / 1805.0;\n\n    float t2 = t * t;\n\n    return t < a\n      ? 7.5625 * t2\n      : t < b\n        ? 9.075 * t2 - 9.9 * t + 3.4\n        : t < c\n          ? ca * t2 - cb * t + cc\n          : 10.8 * t * t - 20.52 * t + 10.72;\n  }\n\n  float bounceIn(float t) {\n    return 1.0 - bounceOut(1.0 - t);\n  }\n\n  float bounceInOut(float t) {\n    return t < 0.5\n      ? 0.5 * (1.0 - bounceOut(1.0 - t * 2.0))\n      : 0.5 * bounceOut(t * 2.0 - 1.0) + 0.5;\n  }\n\n  float circularInOut(float t) {\n    return t < 0.5\n      ? 0.5 * (1.0 - sqrt(1.0 - 4.0 * t * t))\n      : 0.5 * (sqrt((3.0 - 2.0 * t) * (2.0 * t - 1.0)) + 1.0);\n  }\n\n  float circularIn(float t) {\n    return 1.0 - sqrt(1.0 - t * t);\n  }\n\n  float circularOut(float t) {\n    return sqrt((2.0 - t) * t);\n  }\n\n  float cubicInOut(float t) {\n    return t < 0.5\n      ? 4.0 * t * t * t\n      : 0.5 * pow(2.0 * t - 2.0, 3.0) + 1.0;\n  }\n\n  float cubicIn(float t) {\n    return t * t * t;\n  }\n\n  float cubicOut(float t) {\n    float f = t - 1.0;\n    return f * f * f + 1.0;\n  }\n\n  float elasticInOut(float t) {\n    return t < 0.5\n      ? 0.5 * sin(+13.0 * HALF_PI * 2.0 * t) * pow(2.0, 10.0 * (2.0 * t - 1.0))\n      : 0.5 * sin(-13.0 * HALF_PI * ((2.0 * t - 1.0) + 1.0)) * pow(2.0, -10.0 * (2.0 * t - 1.0)) + 1.0;\n  }\n\n  float elasticIn(float t) {\n    return sin(13.0 * t * HALF_PI) * pow(2.0, 10.0 * (t - 1.0));\n  }\n\n  float elasticOut(float t) {\n    return sin(-13.0 * (t + 1.0) * HALF_PI) * pow(2.0, -10.0 * t) + 1.0;\n  }\n\n  float exponentialInOut(float t) {\n    return t == 0.0 || t == 1.0\n      ? t\n      : t < 0.5\n        ? +0.5 * pow(2.0, (20.0 * t) - 10.0)\n        : -0.5 * pow(2.0, 10.0 - (t * 20.0)) + 1.0;\n  }\n\n  float exponentialIn(float t) {\n    return t == 0.0 ? t : pow(2.0, 10.0 * (t - 1.0));\n  }\n\n  float exponentialOut(float t) {\n    return t == 1.0 ? t : 1.0 - pow(2.0, -10.0 * t);\n  }\n\n  float linear(float t) {\n    return t;\n  }\n\n  float quadraticInOut(float t) {\n    float p = 2.0 * t * t;\n    return t < 0.5 ? p : -p + (4.0 * t) - 1.0;\n  }\n\n  float quadraticIn(float t) {\n    return t * t;\n  }\n\n  float quadraticOut(float t) {\n    return -t * (t - 2.0);\n  }\n\n  float quarticInOut(float t) {\n    return t < 0.5\n      ? +8.0 * pow(t, 4.0)\n      : -8.0 * pow(t - 1.0, 4.0) + 1.0;\n  }\n\n  float quarticIn(float t) {\n    return pow(t, 4.0);\n  }\n\n  float quarticOut(float t) {\n    return pow(t - 1.0, 3.0) * (1.0 - t) + 1.0;\n  }\n\n  float qinticInOut(float t) {\n    return t < 0.5\n      ? +16.0 * pow(t, 5.0)\n      : -0.5 * pow(2.0 * t - 2.0, 5.0) + 1.0;\n  }\n\n  float qinticIn(float t) {\n    return pow(t, 5.0);\n  }\n\n  float qinticOut(float t) {\n    return 1.0 - (pow(t - 1.0, 5.0));\n  }\n\n  float sineInOut(float t) {\n    return -0.5 * (cos(PI * t) - 1.0);\n  }\n\n  float sineIn(float t) {\n    return sin((t - 1.0) * HALF_PI) + 1.0;\n  }\n\n  float sineOut(float t) {\n    return sin(t * HALF_PI);\n  }\n";
var easings = createShader({
  vertexHeader: functions,
  fragmentHeader: functions
});
function provideEasingFunctions () {
  return easings;
}

function provideLifetime () {
  var config = {
    lifetime: {
      delay: 0,
      duration: 1
    }
  };
  return createShader({
    attributes: {
      time: {
        type: "vec2"
      }
    },
    varyings: {
      v_progress: {
        type: "float"
      },
      v_age: {
        type: "float"
      }
    },
    vertexMain: "\n      v_age = u_time - time.x;\n      v_progress = v_age / (time.y - time.x);\n\n      if (v_progress < 0.0 || v_progress > 1.0) {\n        csm_Position *= 0.0;;\n      }\n    ",
    fragmentMain: "\n      /* Discard this instance if it is not in the current time range */\n      if (v_progress < 0.0 || v_progress > 1.0) {\n        discard;\n      }\n    ",
    config: config,
    resetConfig: function resetConfig(mesh) {
      config.lifetime.delay = 0;
      config.lifetime.duration = Infinity;
    },
    applyConfig: function applyConfig(mesh, cursor) {
      /* Set times */
      var currentTime = mesh.material.uniforms.u_time.value;
      mesh.geometry.attributes.time.setXY(cursor, currentTime + config.lifetime.delay, currentTime + config.lifetime.delay + config.lifetime.duration);
    }
  });
}

function provideResolution () {
  var u_resolution = {
    type: "vec2",
    value: [window.innerWidth, window.innerHeight]
  };
  return createShader({
    uniforms: {
      u_resolution: u_resolution
    },
    update: function update(_ref) {
      var size = _ref.size;
      u_resolution.value = [size.width, size.height];
    }
  });
}

function provideTime () {
  var u_time = {
    type: "float",
    value: 0
  };
  return createShader({
    uniforms: {
      u_time: u_time
    },
    update: function update(_, dt) {
      return u_time.value += dt;
    }
  });
}

function softParticles () {
  var softness = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;
  var fun = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "clamp(distance / softness, 0.0, 1.0)";
  var uniforms = {
    u_cameraNear: {
      type: "float",
      value: 0
    },
    u_cameraFar: {
      type: "float",
      value: 1
    }
  };
  return createShader({
    uniforms: uniforms,
    varyings: {
      v_viewZ: {
        type: "float"
      }
    },
    vertexMain: "\n      vec4 viewPosition\t= viewMatrix * instanceMatrix * modelMatrix * vec4(csm_Position, 1.0);\n      v_viewZ = viewPosition.z;\n    ",
    fragmentHeader: "\n      float readDepth(vec2 coord) {\n        float depthZ = texture2D(u_depth, coord).x;\n        float viewZ = perspectiveDepthToViewZ(depthZ, u_cameraNear, u_cameraFar);\n        return viewZ;\n      }\n    ",
    fragmentMain: "\n      /* Normalize fragment coordinates to screen space */\n      vec2 screenUv = gl_FragCoord.xy / u_resolution;\n\n      /* Get the existing depth at the fragment position */\n      float depth = readDepth(screenUv);\n\n      {\n        /* Prepare some convenient local variables */\n        float d = depth;\n        float z = v_viewZ;\n        float softness = ".concat(formatValue(softness), ";\n\n        /* Calculate the distance to the fragment */\n        float distance = z - d;\n\n        /* Apply the distance to the fragment alpha */\n        csm_DiffuseColor.a *= ").concat(fun, ";\n      }\n    "),
    update: function update(_ref) {
      var camera = _ref.camera;
      uniforms.u_cameraNear.value = camera.near;
      uniforms.u_cameraFar.value = camera.far;
    }
  });
}

var index = /*#__PURE__*/Object.freeze({
  __proto__: null,
  animateColors: animateColors,
  animateMovement: animateMovement,
  animateScale: animateScale,
  billboarding: billboarding,
  depthTexture: provideDepthTexture,
  provideEasingFunctions: provideEasingFunctions,
  provideLifetime: provideLifetime,
  provideResolution: provideResolution,
  provideTime: provideTime,
  softParticles: softParticles
});

new three.Vector3();
new three.Matrix4();
function useMeshParticles(maxParticles, safetySize) {
  /* The safetySize allows us to emit a batch of particles that would otherwise
     exceed the maximum instance count (which would make WebGL crash.) This way, we don't
     have to upload the entirety of all buffers every time the playhead wraps back to 0. */
  var maxInstanceCount = maxParticles + safetySize;
  var imesh = React.useRef(null);
  React.useLayoutEffect(function () {
    prepareInstancedMesh(imesh.current, imesh.current.material.__vfx.shader, maxInstanceCount);
  }, []);
  var spawnParticle = React.useMemo(function () {
    /* The cursor automatically advances every time a new particle is spawned. */
    var cursor = 0;
    /* This function will spawn new particles. */

    return function (count, setup) {
      /* Grab some stuff we need */
      var _imesh$current = imesh.current,
          material = _imesh$current.material,
          geometry = _imesh$current.geometry;
      var shader = material.__vfx.shader;
      var attributes = geometry.attributes;
      /*
      Safety check: if we've reached the end of the buffers, it means the user picked a safety
      size too small for their use case. We don't want to crash the application, so let's log a
      warning and discard the particle.
      */

      if (cursor + count > maxInstanceCount) {
        console.warn("Spawned too many particles this frame. Discarding. Consider increasing the safetySize.");
        return;
      }
      /* Configure the attributes to upload only the updated parts to the GPU. */


      var allAttributes = [imesh.current.instanceMatrix].concat(_toConsumableArray(Object.values(attributes)));
      allAttributes.forEach(function (attribute) {
        attribute.needsUpdate = true;
        attribute.updateRange.offset = cursor * attribute.itemSize;
        attribute.updateRange.count = count * attribute.itemSize;
      });
      /* Spawn particles, yay! */

      for (var i = 0; i < count; i++) {
        /* Initialize new particle */
        shader.resetConfig(imesh.current);
        setup === null || setup === void 0 ? void 0 : setup(shader.config, i);
        shader.applyConfig(imesh.current, cursor);
        /* Advance playhead */

        cursor++;
      }
      /* Increase count of imesh to match playhead */


      if (cursor > imesh.current.count) {
        imesh.current.count = cursor;
      }
      /* If we've gone past the number of max particles, reset the playhead. */


      if (cursor > maxParticles) {
        cursor = 0;
      }
    };
  }, []);
  return [imesh, {
    spawnParticle: spawnParticle
  }];
}

var _excluded$2 = ["maxParticles", "safetySize", "children", "geometry"];
var MeshParticles = /*#__PURE__*/React.forwardRef(function (_ref, ref) {
  var _ref$maxParticles = _ref.maxParticles,
      maxParticles = _ref$maxParticles === void 0 ? 1000 : _ref$maxParticles,
      _ref$safetySize = _ref.safetySize,
      safetySize = _ref$safetySize === void 0 ? 100 : _ref$safetySize,
      children = _ref.children,
      geometry = _ref.geometry,
      props = _objectWithoutProperties(_ref, _excluded$2);

  var _useMeshParticles = useMeshParticles(maxParticles, safetySize),
      _useMeshParticles2 = _slicedToArray(_useMeshParticles, 2),
      imesh = _useMeshParticles2[0],
      api = _useMeshParticles2[1];

  return /*#__PURE__*/React__default["default"].createElement("instancedMesh", _extends({
    ref: mergeRefs__default["default"]([imesh, ref]),
    args: [geometry, undefined, maxParticles + safetySize]
  }, props), /*#__PURE__*/React__default["default"].createElement(ParticlesContext.Provider, {
    value: api
  }, children));
});

var _excluded$1 = ["billboard", "softness", "scaleFunction", "colorFunction", "softnessFunction", "depthTexture"],
    _excluded2 = ["update"];
var MeshParticlesMaterial = /*#__PURE__*/React.forwardRef(function (_ref, ref) {
  var _ref$billboard = _ref.billboard,
      billboard = _ref$billboard === void 0 ? false : _ref$billboard,
      _ref$softness = _ref.softness,
      softness = _ref$softness === void 0 ? 0 : _ref$softness,
      scaleFunction = _ref.scaleFunction,
      colorFunction = _ref.colorFunction,
      softnessFunction = _ref.softnessFunction,
      depthTexture = _ref.depthTexture,
      props = _objectWithoutProperties(_ref, _excluded$1);

  var material = React.useRef(null);
  var shader = React.useMemo(function () {
    var layers = [provideTime(), provideLifetime(), provideResolution(), provideEasingFunctions(), softness && provideDepthTexture(depthTexture), billboard && billboarding(), animateScale(scaleFunction), animateMovement(), animateColors(colorFunction), softness && softParticles(softness, softnessFunction)].filter(function (l) {
      return l;
    });
    return combineShaders(layers);
  }, []);

  var _useMemo = React.useMemo(function () {
    return compileShader(shader);
  }, [shader]),
      update = _useMemo.update,
      attrs = _objectWithoutProperties(_useMemo, _excluded2);

  React.useLayoutEffect(function () {
    material.current.__vfx = {
      shader: shader
    };
  }, []);
  fiber.useFrame(update);
  return /*#__PURE__*/React__default["default"].createElement(CustomShaderMaterial__default["default"], _extends({
    ref: mergeRefs__default["default"]([material, ref])
  }, attrs, props));
});

var _excluded = ["maxParticles", "safetySize", "children", "geometry"];
var Particles = /*#__PURE__*/React.forwardRef(function (_ref, ref) {
  var _ref$maxParticles = _ref.maxParticles,
      maxParticles = _ref$maxParticles === void 0 ? 1000 : _ref$maxParticles,
      _ref$safetySize = _ref.safetySize,
      safetySize = _ref$safetySize === void 0 ? 100 : _ref$safetySize,
      children = _ref.children,
      geometry = _ref.geometry,
      props = _objectWithoutProperties(_ref, _excluded);

  var maxInstanceCount = maxParticles + safetySize;
  var imesh = React.useRef(null);
  React.useLayoutEffect(function () {
    /* Prepare geometry */
    var geometry = imesh.current.geometry;
    geometry.setAttribute("lifetime", new three.InstancedBufferAttribute(new Float32Array(maxInstanceCount * 2), 2));
    geometry.setAttribute("velocity", new three.InstancedBufferAttribute(new Float32Array(maxInstanceCount * 3), 3));
  }, []);
  var spawnParticle = React.useMemo(function () {
    /* The cursor automatically advances every time a new particle is spawned. */
    var cursor = 0;
    /* This function will spawn new particles. */

    return function (count) {
      console.log("spawning", count);
      /* Grab some stuff we need */

      var _imesh$current = imesh.current,
          material = _imesh$current.material,
          geometry = _imesh$current.geometry;
      var time = material.uniforms.u_time.value;
      var attributes = geometry.attributes;
      /*
      Safety check: if we've reached the end of the buffers, it means the user picked a safety
      size too small for their use case. We don't want to crash the application, so let's log a
      warning and discard the particle.
      */

      if (cursor + count > maxInstanceCount) {
        console.warn("Spawned too many particles this frame. Discarding. Consider increasing the safetySize.");
        return;
      }
      /* Configure the attributes to upload only the updated parts to the GPU. */


      var allAttributes = [imesh.current.instanceMatrix].concat(_toConsumableArray(Object.values(attributes)));
      allAttributes.forEach(function (attribute) {
        attribute.needsUpdate = true;
        attribute.updateRange.offset = cursor * attribute.itemSize;
        attribute.updateRange.count = count * attribute.itemSize;
      });
      /* Spawn particles, yay! */

      for (var i = 0; i < count; i++) {
        var _geometry$attributes$;

        /* Initialize new particle */
        var position = new three.Vector3();
        var quaternion = new three.Quaternion().random();
        var scale = new three.Vector3().setScalar(1);
        imesh.current.setMatrixAt(cursor, new three.Matrix4().compose(position, quaternion, scale));
        /* Configure lifetime */

        geometry.attributes.lifetime.setXY(cursor, time, time + 1.0 + Math.random());
        /* Configure velocity */

        (_geometry$attributes$ = geometry.attributes.velocity).setXYZ.apply(_geometry$attributes$, [cursor].concat(_toConsumableArray(new three.Vector3(Math.random() * 10 - 5, Math.random() * 10, Math.random() * 10 - 5).toArray())));
        /* Advance playhead */


        cursor++;
      }
      /* Increase count of imesh to match playhead */


      if (cursor > imesh.current.count) {
        imesh.current.count = cursor;
      }
      /* If we've gone past the number of max particles, reset the playhead. */


      if (cursor > maxParticles) {
        cursor = 0;
      }
    };
  }, []);
  return /*#__PURE__*/React__default["default"].createElement("instancedMesh", _extends({
    ref: mergeRefs__default["default"]([imesh, ref]),
    args: [geometry, undefined, maxParticles + safetySize]
  }, props), /*#__PURE__*/React__default["default"].createElement(ParticlesContext.Provider, {
    value: {
      spawnParticle: spawnParticle
    }
  }, children));
});

var Delay = function Delay(_ref) {
  var seconds = _ref.seconds,
      onComplete = _ref.onComplete,
      children = _ref.children;
  var ready = useDelay(seconds, onComplete);
  return ready ? /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null, children) : null;
};
var useDelay = function useDelay(seconds, onComplete) {
  var _useState = React.useState(false),
      _useState2 = _slicedToArray(_useState, 2),
      ready = _useState2[0],
      setReady = _useState2[1];

  var timeRemaining = React.useRef(seconds);
  /*
  Hello! You might be wondering why we don't just use `setTimeout` here. We could do exactly that!
  But we're also preparing for a potential future where the deltatime passed into r3f's useFrame
  hook is scaled to something other than 1.0. Using system-level timeouts would entirely sidestep
  all of that.
  */

  fiber.useFrame(function (_, dt) {
    if (ready) return;
    timeRemaining.current -= dt;

    if (timeRemaining.current <= 0) {
      onComplete === null || onComplete === void 0 ? void 0 : onComplete();
      setReady(true);
    }
  });
  return ready;
};

var Lifetime = function Lifetime(_ref) {
  var children = _ref.children,
      seconds = _ref.seconds;
  var ready = useDelay(seconds);
  return ready ? null : /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null, children);
};

var Repeat = function Repeat(_ref) {
  var children = _ref.children,
      _ref$times = _ref.times,
      times = _ref$times === void 0 ? Infinity : _ref$times,
      _ref$interval = _ref.interval,
      interval = _ref$interval === void 0 ? 1 : _ref$interval;

  var _useState = React.useState(1),
      _useState2 = _slicedToArray(_useState, 2),
      iteration = _useState2[0],
      setIteration = _useState2[1];

  return /*#__PURE__*/React__default["default"].createElement(React.Fragment, {
    key: iteration
  }, iteration < times && /*#__PURE__*/React__default["default"].createElement(Delay, {
    seconds: interval,
    onComplete: function onComplete() {
      return setIteration(iteration + 1);
    }
  }), children);
};

exports.Delay = Delay;
exports.Emitter = Emitter;
exports.Lifetime = Lifetime;
exports.MeshParticles = MeshParticles;
exports.MeshParticlesMaterial = MeshParticlesMaterial;
exports.Particles = Particles;
exports.Repeat = Repeat;
exports.addShaders = addShaders;
exports.combineShaders = combineShaders;
exports.compileShader = compileShader;
exports.createShader = createShader;
exports.formatValue = formatValue;
exports.layers = index;
exports.mutateShader = mutateShader;
exports.prepareInstancedMesh = prepareInstancedMesh;
exports.useDelay = useDelay;
exports.useParticles = useParticles;
